@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2008 Edgewall Software
+# Copyright (C) 2004-2009 Edgewall Software
 # Copyright (C) 2004 Daniel Lundin <daniel@edgewall.com>
 # Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
-# Copyright (C) 2006-2007 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2006-2007 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -16,28 +16,11 @@
 #
 # Author: Daniel Lundin <daniel@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
-#         Christian Boos <cboos@neuf.fr>
+#         Christian Boos <cboos@edgewall.org>
 
-"""
-----
-NOTE: for plugin developers
+"""File metadata management.
 
- The Mimeview API is quite complex and many things there are currently
- a bit difficult to work with (e.g. what an actual `content` might be,
- see last paragraph of this docstring).
-
- So this area is mainly in a ''work in progress'' state, which will
- be improved upon in the near future
- (see http://trac.edgewall.org/ticket/3332).
-
- In particular, if you are interested in writing IContentConverter
- and IHTMLPreviewRenderer components, note that those interfaces
- will be merged into a new style IContentConverter.
- Feel free to contribute remarks and suggestions for improvements
- to the corresponding ticket (#3332).
-----
-
-The `trac.mimeview` module centralize the intelligence related to
+The `trac.mimeview` package centralizes the intelligence related to
 file metadata, principally concerning the `type` (MIME type) of the content
 and, if relevant, concerning the text encoding (charset) used by the content.
 
@@ -54,6 +37,22 @@ needed, that's why we avoid to read the file's content when it's not needed.
 The actual `content` to be converted might be a `unicode` object,
 but it can also be the raw byte string (`str`) object, or simply
 an object that can be `read()`.
+
+----
+NOTE: for plugin developers
+
+  The Mimeview API is quite complex and many things there are currently
+  a bit difficult to work with (e.g. what an actual `content` might be,
+  see the last paragraph of this description).
+
+  So this area is mainly in a ''work in progress'' state, which will
+  be improved upon in the near future (see [trac:ticket:3332 #3332]).
+
+  In particular, if you are interested in writing `IContentConverter`
+  and `IHTMLPreviewRenderer` components, note that those interfaces
+  will be merged into a new style `IContentConverter`.
+  Feel free to contribute remarks and suggestions for improvements
+  to the corresponding ticket ([trac:ticket:3332 #3332]).
 """
 
 import re
@@ -67,13 +66,13 @@ from genshi.input import HTMLParser
 from trac.config import IntOption, ListOption, Option
 from trac.core import *
 from trac.resource import Resource
-from trac.util import reversed, sorted, Ranges
-from trac.util.text import to_utf8, to_unicode
-from trac.util.translation import _
+from trac.util import Ranges, content_disposition
+from trac.util.text import exception_to_unicode, to_utf8, to_unicode
+from trac.util.translation import _, tag_
 
 
 __all__ = ['get_mimetype', 'is_binary', 'detect_unicode', 'Mimeview',
-           'content_to_unicode', 'Context']
+           'content_to_unicode', 'ct_mimetype', 'Context']
 
 class Context(object):
     """Rendering context.
@@ -114,7 +113,7 @@ class Context(object):
         :type resource: `Resource`
         :param href: an `Href` object suitable for creating URLs
         :param perm: a `PermissionCache` object used for restricting the
-                     generated output to "authorized" informations only.
+                     generated output to "authorized" information only.
 
         The actual `.perm` attribute of the rendering context will be bound
         to the given `resource` so that fine-grained permission checks will
@@ -126,8 +125,9 @@ class Context(object):
         self.perm = resource and perm and perm(resource) or perm
         self._hints = None
 
+    @classmethod
     def from_request(cls, req, resource=None, id=False, version=False,
-                     absurls=False):
+                     parent=False, absurls=False):
         """Create a rendering context from a request.
 
         The `perm` and `href` properties of the context will be initialized
@@ -157,11 +157,10 @@ class Context(object):
         else:
             href = None
             perm = None
-        self = cls(Resource(resource, id=id, version=version), href=href,
-                   perm=perm)
+        self = cls(Resource(resource, id=id, version=version, parent=parent),
+                   href=href, perm=perm)
         self.req = req
         return self
-    from_request = classmethod(from_request)
 
     def __repr__(self):
         path = []
@@ -172,7 +171,7 @@ class Context(object):
             context = context.parent
         return '<%s %s>' % (type(self).__name__, ' - '.join(reversed(path)))
 
-    def __call__(self, resource=None, id=False, version=False):
+    def __call__(self, resource=None, id=False, version=False, parent=False):
         """Create a nested rendering context.
 
         `self` will be the parent for the new nested context.
@@ -196,7 +195,8 @@ class Context(object):
         True
         """
         if resource:
-            resource = Resource(resource, id=id, version=version)
+            resource = Resource(resource, id=id, version=version,
+                                parent=parent)
         else:
             resource = self.resource
         context = Context(resource, href=self.href, perm=self.perm)
@@ -306,61 +306,70 @@ class Context(object):
 # Some common MIME types and their associated keywords and/or file extensions
 
 KNOWN_MIME_TYPES = {
-    'application/pdf':        ['pdf'],
-    'application/postscript': ['ps'],
-    'application/rtf':        ['rtf'],
-    'application/x-sh':       ['sh'],
-    'application/x-csh':      ['csh'],
-    'application/x-troff':    ['nroff', 'roff', 'troff'],
-    'application/x-yaml':     ['yml', 'yaml'],
+    'application/javascript': 'js',
+    'application/msword':     'doc dot',
+    'application/pdf':        'pdf',
+    'application/postscript': 'ps',
+    'application/rtf':        'rtf',
+    'application/x-sh':       'sh',
+    'application/x-csh':      'csh',
+    'application/x-troff':    'nroff roff troff',
+    'application/x-yaml':     'yml yaml',
     
-    'application/rss+xml':    ['rss'],
-    'application/xsl+xml':    ['xsl'],
-    'application/xslt+xml':   ['xslt'],
+    'application/rss+xml':    'rss',
+    'application/xsl+xml':    'xsl',
+    'application/xslt+xml':   'xslt',
     
-    'image/x-icon':           ['ico'],
-    'image/svg+xml':          ['svg'],
+    'image/x-icon':           'ico',
+    'image/svg+xml':          'svg',
     
-    'model/vrml':             ['vrml', 'wrl'],
+    'model/vrml':             'vrml wrl',
     
-    'text/css':               ['css'],
-    'text/html':              ['html'],
-    'text/plain':             ['txt', 'TXT', 'text', 'README', 'INSTALL',
-                               'AUTHORS', 'COPYING', 'ChangeLog', 'RELEASE'],
-    'text/xml':               ['xml'],
-    'text/x-csrc':            ['c', 'xs'],
-    'text/x-chdr':            ['h'],
-    'text/x-c++src':          ['cc', 'CC', 'cpp', 'C'],
-    'text/x-c++hdr':          ['hh', 'HH', 'hpp', 'H'],
-    'text/x-csharp':          ['cs'],
-    'text/x-diff':            ['diff', 'patch'],
-    'text/x-eiffel':          ['e'],
-    'text/x-elisp':           ['el'],
-    'text/x-fortran':         ['f'],
-    'text/x-haskell':         ['hs'],
-    'text/x-ini':             ['ini', 'cfg'],
-    'text/x-javascript':      ['js'],
-    'text/x-objc':            ['m', 'mm'],
-    'text/x-ocaml':           ['ml', 'mli'],
-    'text/x-makefile':        ['make', 'mk',
-                               'Makefile', 'makefile', 'GNUMakefile'],
-    'text/x-pascal':          ['pas'],
-    'text/x-perl':            ['pl', 'pm', 'PL', 'perl'],
-    'text/x-php':             ['php', 'php3', 'php4'],
-    'text/x-python':          ['py', 'python'],
-    'text/x-pyrex':           ['pyx'],
-    'text/x-ruby':            ['rb', 'ruby'],
-    'text/x-scheme':          ['scm'],
-    'text/x-textile':         ['txtl', 'textile'],
-    'text/x-vba':             ['vb', 'vba', 'bas'],
-    'text/x-verilog':         ['v', 'verilog'],
-    'text/x-vhdl':            ['vhd'],
+    'text/css':               'css',
+    'text/html':              'html htm',
+    'text/plain':             'txt TXT text README INSTALL '
+                              'AUTHORS COPYING ChangeLog RELEASE',
+    'text/xml':               'xml',
+    
+    # see also TEXT_X_TYPES below
+    'text/x-csrc':            'c xs',
+    'text/x-chdr':            'h',
+    'text/x-c++src':          'cc CC cpp C c++ C++',
+    'text/x-c++hdr':          'hh HH hpp H',
+    'text/x-csharp':          'cs c# C#',
+    'text/x-diff':            'patch',
+    'text/x-eiffel':          'e',
+    'text/x-elisp':           'el',
+    'text/x-fortran':         'f',
+    'text/x-haskell':         'hs',
+    'text/x-ini':             'ini cfg',
+    'text/x-objc':            'm mm',
+    'text/x-ocaml':           'ml mli',
+    'text/x-makefile':        'make mk Makefile GNUMakefile',
+    'text/x-pascal':          'pas',
+    'text/x-perl':            'pl pm PL',
+    'text/x-php':             'php3 php4',
+    'text/x-python':          'py',
+    'text/x-pyrex':           'pyx',
+    'text/x-ruby':            'rb',
+    'text/x-scheme':          'scm',
+    'text/x-textile':         'txtl',
+    'text/x-vba':             'vb vba bas',
+    'text/x-verilog':         'v',
+    'text/x-vhdl':            'vhd',
 }
+for t in KNOWN_MIME_TYPES.keys():
+    types = KNOWN_MIME_TYPES[t].split()
+    if t.startswith('text/x-'):
+        types.append(t[len('text/x-'):])
+    KNOWN_MIME_TYPES[t] = types
 
 # extend the above with simple (text/x-<something>: <something>) mappings
 
-for x in ['ada', 'asm', 'asp', 'awk', 'idl', 'inf', 'java', 'ksh', 'lua',
-          'm4', 'mail', 'psp', 'rfc', 'rst', 'sql', 'tcl', 'tex', 'zsh']:
+TEXT_X_TYPES = """
+    ada asm asp awk idl inf java ksh lua m4 mail psp rfc rst sql tcl tex zsh
+"""
+for x in TEXT_X_TYPES.split():
     KNOWN_MIME_TYPES.setdefault('text/x-%s' % x, []).append(x)
 
 
@@ -373,12 +382,12 @@ for t, exts in KNOWN_MIME_TYPES.items():
         MIME_MAP[e] = t
 
 # Simple builtin autodetection from the content using a regexp
-MODE_RE = re.compile(
-    r"#!.+?env (\w+)|"                       # look for shebang with env
-    r"#!(?:[/\w.-_]+/)?(\w+)|"               # look for regular shebang
-    r"-\*-\s*(?:mode:\s*)?([\w+-]+)\s*-\*-|" # look for Emacs' -*- mode -*-
-    r"vim:.*?(?:syntax|filetype|ft)=(\w+)"   # look for VIM's syntax=<n>
-    )
+MODE_RE = re.compile(r"""
+      \#!.+?env\s+(\w+)                     # 1. look for shebang with env
+    | \#!(?:[/\w.-_]+/)?(\w+)               # 2. look for regular shebang
+    | -\*-\s*(?:mode:\s*)?([\w+-]+)\s*-\*-  # 3. look for Emacs' -*- mode -*-
+    | vim:.*?(?:syntax|filetype|ft)=(\w+)   # 4. look for VIM's syntax=<n>
+    """, re.VERBOSE)
 
 def get_mimetype(filename, content=None, mime_map=MIME_MAP):
     """Guess the most probable MIME type of a file with the given name.
@@ -414,6 +423,10 @@ def get_mimetype(filename, content=None, mime_map=MIME_MAP):
                     return 'application/octet-stream'
         return mimetype
 
+def ct_mimetype(content_type):
+    """Return the mimetype part of a content type."""
+    return (content_type or '').split(';')[0].strip()
+
 def is_binary(data):
     """Detect binary content by checking the first thousand bytes for zeroes.
 
@@ -424,7 +437,7 @@ def is_binary(data):
     return '\0' in data[:1000]
 
 def detect_unicode(data):
-    """Detect different unicode charsets by looking for BOMs (Byte Order Marks).
+    """Detect different unicode charsets by looking for BOMs (Byte Order Mark).
 
     Operate obviously only on `str` objects.
     """
@@ -438,11 +451,25 @@ def detect_unicode(data):
         return None
 
 def content_to_unicode(env, content, mimetype):
-    """Retrieve an `unicode` object from a `content` to be previewed"""
+    """Retrieve an `unicode` object from a `content` to be previewed.
+
+    In case the raw content had an unicode BOM, we remove it.
+
+    >>> from trac.test import EnvironmentStub
+    >>> env = EnvironmentStub()
+    >>> content_to_unicode(env, u"\ufeffNo BOM! h\u00e9 !", '')
+    u'No BOM! h\\xe9 !'
+    >>> content_to_unicode(env, "\xef\xbb\xbfNo BOM! h\xc3\xa9 !", '')
+    u'No BOM! h\\xe9 !'
+
+    """
     mimeview = Mimeview(env)
     if hasattr(content, 'read'):
         content = content.read(mimeview.max_preview_size)
-    return mimeview.to_unicode(content, mimetype)
+    u = mimeview.to_unicode(content, mimetype)
+    if u and u[0] == u'\ufeff':
+        u = u[1:]
+    return u
 
 
 class IHTMLPreviewRenderer(Interface):
@@ -525,8 +552,8 @@ class IHTMLPreviewAnnotator(Interface):
         `context` is the context corresponding to the content being annotated,
         `row` is the tr Element being built, `number` is the line number being
         processed and `line` is the line's actual content.
-        `annotations` is whatever additional data the `get_annotation_data`
-        method decided to provide.
+        `data` is whatever additional data the `get_annotation_data` method
+        decided to provide.
         """
 
 
@@ -568,8 +595,29 @@ class IContentConverter(Interface):
         output_mime_type) or None if conversion is not possible."""
 
 
+class Content(object):
+    """A lazy file-like object that only reads `input` if necessary."""
+    def __init__(self, input, max_size):
+        self.input = input
+        self.max_size = max_size
+        self.content = None
+    
+    def read(self, size=-1):
+        if size == 0:
+            return ''
+        if self.content is None:
+            self.content = StringIO(self.input.read(self.max_size))
+        return self.content.read(size)
+    
+    def reset(self):
+        if self.content is not None:
+            self.content.seek(0)
+
+
 class Mimeview(Component):
-    """A generic class to prettify data, typically source code."""
+    """Generic HTML renderer for data, typically source code."""
+
+    required = True
 
     renderers = ExtensionPoint(IHTMLPreviewRenderer)
     annotators = ExtensionPoint(IHTMLPreviewAnnotator)
@@ -579,17 +627,23 @@ class Mimeview(Component):
         """Charset to be used when in doubt.""")
 
     tab_width = IntOption('mimeviewer', 'tab_width', 8,
-        """Displayed tab width in file preview (''since 0.9'').""")
+        """Displayed tab width in file preview. (''since 0.9'')""")
 
     max_preview_size = IntOption('mimeviewer', 'max_preview_size', 262144,
-        """Maximum file size for HTML preview. (''since 0.9'').""")
+        """Maximum file size for HTML preview. (''since 0.9'')""")
 
     mime_map = ListOption('mimeviewer', 'mime_map',
         'text/x-dylan:dylan,text/x-idl:ice,text/x-ada:ads:adb',
         doc="""List of additional MIME types and keyword mappings.
         Mappings are comma-separated, and for each MIME type,
         there's a colon (":") separated list of associated keywords
-        or file extensions. (''since 0.10'').""")
+        or file extensions. (''since 0.10'')""")
+
+    treat_as_binary = ListOption('mimeviewer', 'treat_as_binary',
+        'application/octet-stream,application/pdf,application/postscript,'
+        'application/msword,application/rtf,',
+        doc="""Comma-separated list of MIME types that should be treated as
+        binary data. (''since 0.11.5'')""")
 
     def __init__(self):
         self._mime_map = None
@@ -602,7 +656,8 @@ class Mimeview(Component):
         component appended. Output is ordered from best to worst quality."""
         converters = []
         for converter in self.converters:
-            for k, n, e, im, om, q in converter.get_supported_conversions():
+            conversions = converter.get_supported_conversions() or []
+            for k, n, e, im, om, q in conversions:
                 if im == mimetype and q > 0:
                     converters.append((k, n, e, im, om, q, converter))
         converters = sorted(converters, key=lambda i: i[-2], reverse=True)
@@ -623,26 +678,27 @@ class Mimeview(Component):
                 content = content.read(self.max_preview_size)
             full_mimetype = self.get_mimetype(filename, content)
         if full_mimetype:
-            mimetype = full_mimetype.split(';')[0].strip() # split off charset
+            mimetype = ct_mimetype(full_mimetype)   # split off charset
         else:
             mimetype = full_mimetype = 'text/plain' # fallback if not binary
 
         # Choose best converter
-        candidates = list(self.get_supported_conversions(mimetype))
+        candidates = list(self.get_supported_conversions(mimetype) or [])
         candidates = [c for c in candidates if key in (c[0], c[4])]
         if not candidates:
-            raise TracError(_('No available MIME conversions from %(old)s '
-                              'to %(new)s', old=mimetype, new=key))
+            raise TracError(
+                _("No available MIME conversions from %(old)s to %(new)s",
+                  old=mimetype, new=key))
 
         # First successful conversion wins
         for ck, name, ext, input_mimettype, output_mimetype, quality, \
                 converter in candidates:
             output = converter.convert_content(req, mimetype, content, ck)
-            if not output:
-                continue
-            return (output[0], output[1], ext)
-        raise TracError(_('No available MIME conversions from %(old)s to '
-                          '%(new)s', old=mimetype, new=key))
+            if output:
+                return (output[0], output[1], ext)
+        raise TracError(
+            _("No available MIME conversions from %(old)s to %(new)s",
+              old=mimetype, new=key))
 
     def get_annotation_types(self):
         """Generator that returns all available annotation types."""
@@ -662,6 +718,10 @@ class Mimeview(Component):
         content.
 
         Return a string containing the XHTML text.
+        
+        When rendering with an `IHTMLPreviewRenderer` fails, a warning is added
+        to the request associated with the context (if any), unless the
+        `disable_warnings` hint is set to `True`.
         """
         if not content:
             return ''
@@ -677,7 +737,7 @@ class Mimeview(Component):
                 content = content.read(self.max_preview_size)
             full_mimetype = self.get_mimetype(filename, content)
         if full_mimetype:
-            mimetype = full_mimetype.split(';')[0].strip() # split off charset
+            mimetype = ct_mimetype(full_mimetype)   # split off charset
         else:
             mimetype = full_mimetype = 'text/plain' # fallback if not binary
 
@@ -687,20 +747,25 @@ class Mimeview(Component):
             qr = renderer.get_quality_ratio(mimetype)
             if qr > 0:
                 candidates.append((qr, renderer))
-        candidates.sort(lambda x,y: cmp(y[0], x[0]))
+        candidates.sort(lambda x, y: cmp(y[0], x[0]))
+        
+        # Wrap file-like object so that it can be read multiple times
+        if hasattr(content, 'read'):
+            content = Content(content, self.max_preview_size)
 
         # First candidate which renders successfully wins.
         # Also, we don't want to expand tabs more than once.
         expanded_content = None
-        errors = []
         for qr, renderer in candidates:
             if force_source and not getattr(renderer, 'returns_source', False):
                 continue # skip non-source renderers in force_source mode
+            if isinstance(content, Content):
+                content.reset()
             try:
                 ann_names = annotations and ', '.join(annotations) or \
-                           'No annotations'
-                self.log.debug('Trying to render HTML preview using %s [%s]'
-                               % (renderer.__class__.__name__, ann_names))
+                           'no annotations'
+                self.log.debug('Trying to render HTML preview using %s [%s]',
+                               renderer.__class__.__name__, ann_names)
 
                 # check if we need to perform a tab expansion
                 rendered_content = content
@@ -739,10 +804,15 @@ class Mimeview(Component):
                     return tag.div(class_='code')(tag.pre(result)).generate()
 
             except Exception, e:
-                self.log.warning('HTML preview using %s failed (%s)' %
-                                 (renderer, e), exc_info=True)
-                errors.append((renderer, e))
-        return errors
+                self.log.warning('HTML preview using %s failed: %s',
+                                 renderer.__class__.__name__,
+                                 exception_to_unicode(e, traceback=True))
+                if context.req and not context.get_hint('disable_warnings'):
+                    from trac.web.chrome import add_warning
+                    add_warning(context.req,
+                        _("HTML preview using %(renderer)s failed (%(err)s)",
+                          renderer=renderer.__class__.__name__,
+                          err=exception_to_unicode(e)))
 
     def _render_source(self, context, stream, annotations, marks=None):
         from trac.web.chrome import add_warning
@@ -753,6 +823,7 @@ class Mimeview(Component):
                 labels[atype] = alabel
                 titles[atype] = atitle
                 annotators[atype] = annotator
+        annotations = [a for a in annotations if a in annotators]
 
         if isinstance(stream, list):
             stream = HTMLParser(StringIO('\n'.join(stream)))
@@ -769,11 +840,10 @@ class Mimeview(Component):
             try:
                 data = (annotator, annotator.get_annotation_data(context))
             except TracError, e:
-                self.log.warning("Can't use annotator '%s': %s" %
-                                 (a, e.message))
-                add_warning(context.req, tag.strong("Can't use ", tag.em(a),
-                                               " annotator:") +
-                                    tag.pre(e.message))
+                self.log.warning("Can't use annotator '%s': %s", a, e.message)
+                add_warning(context.req, tag.strong(
+                    tag_("Can't use %(annotator)s annotator: %(error)s",
+                         annotator=tag.em(a), error=tag.pre(e.message))))
                 data = (None, None)
             annotator_datas.append(data)
 
@@ -826,6 +896,18 @@ class Mimeview(Component):
                 return utf
         return self.default_charset
 
+    @property
+    def mime_map(self):
+        # Extend default extension to MIME type mappings with configured ones
+        if not self._mime_map:
+            self._mime_map = MIME_MAP.copy()
+            for mapping in self.config['mimeviewer'].getlist('mime_map'):
+                if ':' in mapping:
+                    assocations = mapping.split(':')
+                    for keyword in assocations: # Note: [0] kept on purpose
+                        self._mime_map[keyword] = assocations[0]
+        return self._mime_map
+
     def get_mimetype(self, filename, content=None):
         """Infer the MIME type from the `filename` or the `content`.
 
@@ -835,16 +917,8 @@ class Mimeview(Component):
         charset information (i.e. "<mimetype>; charset=..."),
         or `None` if detection failed.
         """
-        # Extend default extension to MIME type mappings with configured ones
-        if not self._mime_map:
-            self._mime_map = MIME_MAP
-            for mapping in self.config['mimeviewer'].getlist('mime_map'):
-                if ':' in mapping:
-                    assocations = mapping.split(':')
-                    for keyword in assocations: # Note: [0] kept on purpose
-                        self._mime_map[keyword] = assocations[0]
 
-        mimetype = get_mimetype(filename, content, self._mime_map)
+        mimetype = get_mimetype(filename, content, self.mime_map)
         charset = None
         if mimetype:
             charset = self.get_charset(content, mimetype)
@@ -852,6 +926,18 @@ class Mimeview(Component):
             mimetype += '; charset=' + charset
         return mimetype
 
+    def is_binary(self, mimetype=None, filename=None, content=None):
+        """Check if a file must be considered as binary."""
+        if not mimetype and filename:
+            mimetype = self.get_mimetype(filename, content)
+        if mimetype:
+            mimetype = ct_mimetype(mimetype)
+            if mimetype in self.treat_as_binary:
+                return True
+        if content is not None and is_binary(content):
+            return True
+        return False
+
     def to_utf8(self, content, mimetype=None):
         """Convert an encoded `content` to utf-8.
 
@@ -880,7 +966,7 @@ class Mimeview(Component):
                 types[mimetype] = (mode, int(quality))
             except (TypeError, ValueError):
                 self.log.warning("Invalid mapping '%s' specified in '%s' "
-                                 "option." % (mapping, option))
+                                 "option.", mapping, option)
         return types
     
     def preview_data(self, context, content, length, mimetype, filename,
@@ -899,24 +985,25 @@ class Mimeview(Component):
         else:
             result = self.render(context, mimetype, content, filename, url,
                                  annotations, force_source=force_source)
-            if isinstance(result, list):
-                data['errors'] = result
-            else:
-                data['rendered'] = result
+            data['rendered'] = result
         return data
 
     def send_converted(self, req, in_type, content, selector, filename='file'):
         """Helper method for converting `content` and sending it directly.
 
         `selector` can be either a key or a MIME Type."""
-        from trac.web import RequestDone
+        from trac.web.api import RequestDone
         content, output_type, ext = self.convert_content(req, in_type,
                                                          content, selector)
+        if isinstance(content, unicode):
+            content = content.encode('utf-8')
         req.send_response(200)
         req.send_header('Content-Type', output_type)
+        req.send_header('Content-Length', len(content))
         if filename:
-            req.send_header('Content-Disposition', 'filename=%s.%s' % 
-                        (filename, ext))
+            req.send_header('Content-Disposition',
+                            content_disposition('attachment',
+                                                '%s.%s' % (filename, ext)))
         req.end_headers()
         req.write(content)
         raise RequestDone
@@ -1006,7 +1093,7 @@ class LineNumberAnnotator(Component):
     # ITextAnnotator methods
 
     def get_annotation_type(self):
-        return 'lineno', 'Line', 'Line numbers'
+        return 'lineno', _('Line'), _('Line numbers')
 
     def get_annotation_data(self, context):
         return None
@@ -1028,14 +1115,8 @@ class PlainTextRenderer(Component):
     expand_tabs = True
     returns_source = True
 
-    TREAT_AS_BINARY = [
-        'application/pdf',
-        'application/postscript',
-        'application/rtf'
-    ]
-
     def get_quality_ratio(self, mimetype):
-        if mimetype in self.TREAT_AS_BINARY:
+        if mimetype in Mimeview(self.env).treat_as_binary:
             return 0
         return 1
 
@@ -1049,7 +1130,10 @@ class PlainTextRenderer(Component):
 
 
 class ImageRenderer(Component):
-    """Inline image display. Here we don't need the `content` at all."""
+    """Inline image display.
+    
+    This component doesn't need the `content` at all.
+    """
     implements(IHTMLPreviewRenderer)
 
     def get_quality_ratio(self, mimetype):
@@ -1064,7 +1148,7 @@ class ImageRenderer(Component):
 
 
 class WikiTextRenderer(Component):
-    """Render files containing Trac's own Wiki formatting markup."""
+    """HTML renderer for files containing Trac's own Wiki formatting markup."""
     implements(IHTMLPreviewRenderer)
 
     def get_quality_ratio(self, mimetype):
